{ Pdf-Cmp Copyright 2000 by Stefan Menten
  Pdf-File format is Copyright by Adobe Systems

  Unit UPdfCmp
  main unit for generating pdf-files

  Pdf-Cmp is a component to generate Pdf-files from within
  Delphi Applications

  It is not intended to be a Pdf-Editor but is usefull for creating
  files from a structured source like data from a database

  needed files:

  UPdfBaseFonts
  UPdfFilters
  UPdfImages
  UPdfDefinition

  xProcs (Copyright (c) 1995 Stefan Bther, stefc@fabula.com)

  }



unit UPdfCmp;

{$R UPdfCmp.DCR}

interface

Uses Windows, Messages, SysUtils, Classes, Graphics,
     UPdfBaseFonts, UPdfFilters, UPdfImages, UPdfDefinition;

// objects -------------------------------------------------------------------

type
     TPageSettings = Class(TPersistent)
       private
         FPaper: TPaper;
         FMediaBox: TMediaBox;
         FCropBox: TMediaBox;
         FPageMargins: TPageMargins;
         FOrientation: TPdfOrientation;
         procedure SetPdfPage;
         procedure SetPaper(Value: TPaper);
         procedure SetMargins(Value: TPageMargins);
         procedure SetOrientation(Value: TPdfOrientation);
       public
         constructor create;
       published
         procedure Margins(const maLeft, maRight, maTop, maBottom: Integer);
         property PageSize: TPaper read FPaper write SetPaper default A4;
//         property Margins: TPageMargins read FPageMargins write SetMargins;
         property Orientation: TPdfOrientation
                  read FOrientation write SetOrientation;
       end;

type
     TPdfCmp = Class;

     // general object for Pdf
     TPdfObject = class
       PdfOwner: TPdfCmp;
       ObjType: String;
       ObjID: Word;
       Generation: Word;
     public
       constructor Create;
     end;

     // XObject for Bitmaps
     TPdfXImage = Class(TPdfObject)
       Name: String;
       Width, Height: Integer;
       Left, Top: Double;
       BitsPerComponent: Byte;
       Filter: TFilter;
       ColorSpace: String;
       ColorCount: Byte;
       ImageStream: String;
     public
       procedure SetBitmap(const Left, Top: double;
                 const BitmapPath: String);
       procedure Write(var StrList: TStringList; const Nr: Integer);
     end;

     // Pdf-Header
     TPdfHeader = Class(TPdfObject)
       PdfVersion: String;
       procedure Write(var StrList: TStringList;
                const Binary: Boolean; const Version: String);
     end;

     // Pdf-Body
     // the root note of the document
     TPdfCatalog = Class(TPdfObject)
       Pages: Integer;
       Outlines: Integer;
       procedure Write(var StrList: TStringList;
                 const sPages, sOutline: String);
     end;
     // bookmark dictionary, for future use
     TPdfOutlines = Class(TPdfObject)
       OutlCount: Integer;
       procedure Write(var StrList: TStringList);
     end;
     // Content for one page
     TPdfContents = Class(TPdfObject)
         BmpCount: Integer;
         ConList: TStringList;
         Bitmaps: array[1..10] of TPdfXImage;
       public
         constructor Create;
         procedure Write(var StrList: TStringList);
       end;

     // description for one page in document
     TPdfPage = Class(TPdfObject)
       MediaBox:  TMediaBox;
       Parent:    Word;          // reference to pages
       Resources: String;           // like fonts, ProcSet aso
       PdfContents:  TPdfContents;      // reference to page description
       CropBox:   TMediaBox;
     public
       constructor Create;
       procedure Write(var StrList: TStringList);
     end;

     // Note for pages in document
     TPdfPages = Class(TPdfObject)
       IsRoot: Boolean;
       Parent: Word;
       PageCount:  Word;
       Kids:   array[1..c_MaxPageCount] of Word;
     public
       procedure Write(var StrList: TStringList; const aPages: array of TPdfPage);
     end;

     // ProcRef
     TPdfProcRef = Class(TPdfObject)
       procedure Write(var StrList: TStringList);
     end;

     TPdfBody = Class(TPdfObject)
       PdfCatalog: TPdfCatalog;
       PdfOutlines: TPdfOutlines;
       PdfPages: TPdfPages;
       PdfPage: array[1..c_MaxPageCount] of TPdfPage;
       PdfProcRef: TPdfProcRef;
     public
       procedure InitBody;
       function WritePdfBody(var StrList: TStringList;
                const iPageCount: Integer): Boolean;
     end;
     // Pdf-Crossreferenc
     TPdfCrossRefernec = Class(TPdfObject)
       XRefCount: Word;
       StartOffset: Word;
       procedure Write(var StrList: TStringList);
     end;

     // Pdf-Trailer
     TPdfTrailer = Class(TPdfObject)
       procedure Write(var StrList: TStringList);
     end;

     // the PDF-Generator component ------------------------------------------
     TPdfCmp = Class(TComponent)
     private
       // private definitions
       FYPosActual: double;
       FPageCount: Word;
       FPageActual: Word;
       // the objects a Pdf-File is made from
       FPdfHeader: TPdfHeader;       // the Pdf header
       FPdfBody: TPdfBody;           // the Pdf body
       FPdfActPage: ^TPdfPage;       // the actual page to paint to
       FPdfCrossRefernec: TPdfCrossRefernec;
       FPdfTrailer: TPdfTrailer;
       FFileNamePdf: String;         // File name to write Pdf to
       FPdfVersion: String;          // Version to save Pdf-file
       FPdfStrList: TStringList;     // String list that holds file contens
       FPageSettings: TPageSettings;
       PdfBaseFonts: TPdfBaseFonts;
       procedure SetFileNamePdf(const Value: String);    // Set the file name
       procedure SetPdfVersion(const Value: String);     // Set the version
       procedure SetDefaults;                            // Set defaults to Records
       // get X position depended to align
       function GetXPosition(const sText: String;
                taAlign: TPdfAlignment; const iFont, iFontSize: Integer): double;
     protected
       // protected definitions

       BaseFonts: array[1..14] of String; // holds references to base fonts
       procedure PdfInit;
       procedure WriteFonts(var StrList: TStringList);
       function SplitString(const sString: String): String;
       function WritePdfFile: Boolean;      // write the entire file to disk
       // add floating text to contents
       function AddFloatingText(const xStart, yStart, MaxWidth: double;
                const PdfAlign: TPdfAlignment; sText: String): double;
     public
       // public definitions
       PdfWriteParDef: TPdfWriteParDef;
       constructor Create(AOwner : TComponent); override;
       destructor Destroy; override;
       procedure SetMargins(const maLeft, maRight, maTop, maBottom: Integer);
       property PdfPageActual: Word read FPageActual write FPageActual;  // actual page
       property PdfPageCount: Word read FPageCount write FPageCount;     // total count of pages
       procedure SetOrientation(const Orientation: TPdfOrientation);
       function GetTextLengthMM(const byFontNr, byFontSize: Byte; const sText: String): Word;
       function PageAppend: Boolean;       // adds a new page at the end
       // add text to actual page
       procedure AddText(const xPos, yPos: Double; const FontPdfName: String;
                 const FontSize: Byte; const PdfText: String);
       // write a complete paragraph
       procedure WriteParagraph(const Heading, Text: String);
       // add bitmap to actual page
       procedure AddBitmap(const Left, Top: double;
                 const BitmapPath: String);
       procedure DrawLineHor(const xPos, yPos, xLen, iLineWidth: Integer);
       procedure DrawLineVer(const xPos, yPos, yLen, iLineWidth: Integer);
       procedure DrawLine(const xTop, yTop, xBot, yBot, iLineWidth: Integer);
       procedure DrawRect(const xTop, yTop, xBot, yBot, iLineWidth: Integer);
       procedure PdfSave; // saves the document to disc
       property ActualY: double read FYPosActual;
     published
       // published definitions
       property PdfVersion: String read FPdfVersion write SetPdfVersion;
       property PdfFileName: String read FFileNamePdf write SetFileNamePdf;
       property PageSettings: TPageSettings read FPageSettings write FPageSettings;
     end;

Procedure Register;

implementation

uses xProcs1;

var
   g_IDNextFree: Word;
   g_XREFList: TStringList;
// global to unit

// --- GetNextID ---------------------------------------
// 11.01.2000 21:34:57
// generat next ID for Pdf-object
function GetNextID: Word;
begin
result := g_IDNextFree;
Inc(g_IDNextFree);
end;

// --- GetReferenceTo ---------------------------------------
// 11.01.2000 21:35:22
// return the reference to an existing Pdf-object
function GetReferenceTo(const oPdfObject: TPdfObject): String;
begin
if oPdfObject <> NIL then
Result := IntToStr(TPdfObject(oPdfObject).ObjID) +
       ' ' + IntToStr(TPdfObject(oPdfObject).Generation) + ' R'
end;

// --- WriteObjHeader ---------------------------------------
// 11.01.2000 21:35:59
// Write the Pdf-Header of the object to the String-list of the Pdf-File
//
procedure WriteObjHeader(var sList: TStringList; const oPdfObject: TPdfObject);
begin
with sList do
  begin
  // add Object-ID, Generation and 'obj' as type
  Add(IntToStr(TPdfObject(oPdfObject).ObjID) +
      ' ' + IntToStr(TPdfObject(oPdfObject).Generation) + ' obj');
  // begin the object
  Add('<<');
  // set the type
  Add('/Type /' + TPdfObject(oPdfObject).ObjType);
  end;
end;

// --- GetByteOffset ---------------------------------------
// 11.01.2000 21:37:58
// get the entiry length of the file so far
//
function GetByteOffset(const sList: TStringList): Word;
var
   x: Integer;
begin
Result := 0;
// length of all strings + 2 for CR and LF
for x := 0 to sList.Count -1 do Result := Result + Length(sList[x]);

//Result := Result  + Trunc(sList.Count*2)-1;
//result := Length(sList.Text);
end;

// --- AddObj2XREF ---------------------------------------
// 11.01.2000 21:38:55
// Add an object reference to the list of objects
//
procedure AddObj2XREF(const wOffset, Generation: Word; const sState: String);
begin
g_XREFList.Add(Copy('0000000000', 1, 10-Length(IntToStr(wOffset))) + IntToStr(wOffset) + ' ' +
               Copy('00000', 1, 5-Length(IntToStr(Generation))) + IntToStr(Generation) + ' ' +
               sState);
end;

// --- TPageSettings ---------------------------------------
// 11.01.2000 21:39:21
// funtions and methods of the object
// TPageSettings holds all settings of the page used for pdf-file
//

// --- TPageSettings.create ---------------------------------------
// 11.01.2000 21:40:30
// create and init the object
// set margins to default = 10 mm
constructor TPageSettings.create;
begin
inherited create;
with FPageMargins do
  begin
  Left   := 10;
  Right  := 10;
  Top    := 10;
  Bottom := 10;
  end;
SetPdfPage;
end;

// --- TPageSettings.SetPaper ---------------------------------------
// 11.01.2000 21:41:17
// set paper assosiated to pager list
procedure TPageSettings.SetPaper(Value: TPaper);
begin
FPaper := Value;
SetPdfPage;
end;

// --- TPageSettings.SetMargins ---------------------------------------
// 11.01.2000 21:41:27
// Set margins of page to CropBox
procedure TPageSettings.SetMargins(Value: TPageMargins);
begin
FPageMargins := Value;
SetPdfPage;
end;

// --- TPageSettings.SetPdfPage ---------------------------------------
// 11.01.2000 21:41:46
// set values for pdf from page size and margins
procedure TPageSettings.SetPdfPage;

  procedure SetMediaBox(const Width, height: Integer; Orientation: TPdfOrientation);
  begin
  with FMediaBox do
    begin
    mbTop := 0;
    mbLeft := 0;
    if Orientation = PdfLandscape then
       begin
       mbHeight := (Width / cMM2Inch) * cInch2UserSpace;
       mbWidth  := (Height / cMM2Inch) * cInch2UserSpace;
       end
    else
       begin
       mbWidth  := (Width / cMM2Inch) * cInch2UserSpace;
       mbHeight := (Height / cMM2Inch) * cInch2UserSpace;

       end;
    end;
  end;

begin
// set size of page
case FPaper of
  A3:           begin
                SetMediaBox(297, 420, Self.FOrientation);
                end;
  A4:           begin
                SetMediaBox(210, 297, Self.FOrientation);
                end;
  A5:           begin
                SetMediaBox(148, 210, Self.FOrientation);
                end;
  Letter:       begin
                SetMediaBox(215, 279, Self.FOrientation);
                end;
  Legal:        begin
                SetMediaBox(215, 305, Self.FOrientation);
                end;
  DL_Envelope:  begin
                SetMediaBox(220, 110, Self.FOrientation);
                end;
  else
  end;
// Set CropBox relativ to MediaBox
with FCropBox, FPageMargins do
  begin
  mbLeft   := Left;
  mbTop    := Top;
  mbHeight := FMediaBox.mbHeight - Top - Bottom;
  mbWidth  := FMediaBox.mbWidth - Left - Right;
  end;
end;

// --- TPageSettings.SetOrientation ---------------------------------------
// 11.01.2000 21:42:27
// set orientation of page
procedure TPageSettings.SetOrientation(Value: TPdfOrientation);
begin
if Value = PdfLandscape then
   FOrientation := PdfLandscape
else
   FOrientation := PdfPortrait;
SetPdfPage;
end;

// --- TPageSettings.Margins ---------------------------------------
// 11.01.2000 21:42:36
// set margins to margins in UserSpace from mm
//
procedure TPageSettings.Margins(const maLeft, maRight, maTop, maBottom: Integer);
begin
// Set default margins
with FPageMargins do
  begin
  Left   := GetUserSpaceFromMM(maLeft);
  Right  := GetUserSpaceFromMM(maRight);
  Top    := GetUserSpaceFromMM(maTop);
  Bottom := GetUserSpaceFromMM(maBottom);
  end;
SetPdfPage;
end;

// --- TPdfObject ---------------------------------------
// 11.01.2000 21:43:14
// TPdfObject is the mother of all objects used in Pdf-Cmd
//

// --- TPdfObject.Create ---------------------------------------
// 11.01.2000 21:44:39
// create the object and get the ID
//
constructor TPdfObject.Create;
begin
inherited Create;
Generation := 0;
ObjID := GetNextID;
end;

// --- TPdfXImage ---------------------------------------
// 11.01.2000 21:45:14
// object for handling images
//

// --- TPdfXImage.SetBitmap ---------------------------------------
// 11.01.2000 21:45:34
// set the bitmap and load it to a stream
//
procedure TPdfXImage.SetBitmap(const Left, Top: double;
          const BitmapPath: String);
var
   tmpBitmap: TPicture;
begin
tmpBitmap := NIL;
try
// create TPicture and load bitmap from file
tmpBitmap := TPicture.create;
tmpBitmap.LoadFromFile(BitmapPath);
// set width and height, Left and top
Self.Width := tmpBitmap.Width;
Self.Height := tmpBitmap.Height;
Self.Left := Left;
Self.Top := PdfOwner.FPageSettings.FMediaBox.mbHeight - Top;
// set BitsPerComponent
// ColorCount sets the max Colors from 0
case tmpBitmap.Bitmap.PixelFormat of
  pf1bit: begin
          BitsPerComponent := 1;
          Self.ColorCount  := 1;
          Self.ColorSpace := EncodeBase85Dec(Self.ColorSpace);
          end;
  pf4bit: begin
          BitsPerComponent := 4;
          Self.ColorCount  := 15;
          Self.ColorSpace := EncodeBase85Dec(c_ColorSpace16);
          end;
  pf8bit: begin
          BitsPerComponent := 8;
          Self.ColorCount  := 255;
          Self.ColorSpace := EncodeBase85Dec(Self.ColorSpace);
          end;
  else
    BitsPerComponent := 0;
  end;
// set the image data and encode to Ascii85
Self.ImageStream := EncodeBase85DecFromStream(GetBitmapData(tmpBitmap.Bitmap));
finally
tmpBitmap.Free;
end;
end;

// --- TPdfXImage.Write ---------------------------------------
// 11.01.2000 21:46:22
// write the bitmap out to the string-list that holds the pdf-file
//
procedure TPdfXImage.Write(var StrList: TStringList; const Nr: Integer);
var
   FilterObjectNr, sTemp: String;
   x, wOffset: Integer;
begin
FilterObjectNr := IntToStr(GetNextID) + ' 0 ';
with StrList do
  begin
  // write image data
  wOffset := GetByteOffset(StrList);
  AddObj2XREF(wOffset, Generation, 'n');
  Add(IntToStr(Self.ObjID) + ' 0 obj');
  Add('<<');
  Add('/Type /XObject');
  Add('/Subtype /Image');
  Add('/Name /Im' + IntToStr(Nr));
  Add('/Filter /ASCII85Decode');
  Add('/Width ' + IntToStr(Self.Width));
  Add('/Height ' + IntToStr(Self.Height));
  Add('/BitsPerComponent ' + IntToStr(Self.BitsPerComponent));
  Add('/ColorSpace [ /Indexed /DeviceRGB ' + IntToStr(Self.ColorCount) +
      ' ' + FilterObjectNr + 'R ]');
  Add('/Length ' + IntToStr(Length(Self.ImageStream)));
  Add('>>');
  Add('stream');
  sTemp := '';
  // write data in pieces of 100 chars
  for x := 1 to Length(Self.ImageStream) do
    begin
    sTemp := sTemp + Self.ImageStream[x];
    if (Length(sTemp) MOD 65) = 0 then
       begin
       Add(sTemp);
       sTemp := '';
       end;
    end;
  // write remaining chars
  if sTemp > '' then Add(sTemp);
  Add('endstream');
  Add('endobj');
  // write filter data
  wOffset := GetByteOffset(StrList);
  AddObj2XREF(wOffset, Generation, 'n');
  Add(FilterObjectNr + 'obj');
  Add('<<');
  Add('/Filter /ASCII85Decode');
  Add('/Length ' + IntToStr(Length(Self.ColorSpace)));
  Add('>>');
  Add('stream');
  Add(Self.ColorSpace);
  Add('endstream');
  Add('endobj');
  end;
end;

// --- TPdfHeader.Write ---------------------------------------
// 11.01.2000 21:47:03
// holds the header information of the pdf-file
//

// --- TPdfHeader.Write ---------------------------------------
// 11.01.2000 21:47:23
// write out the header information to the string-list that holds
// the pdf-file
procedure TPdfHeader.Write(var StrList: TStringList;
         const Binary: Boolean; const Version: String);
begin
// write PDF version
StrList.Add('%PDF-1.2');
StrList.Add('%' + #226#227#207#211);
StrList.Add('  ');
// if Version 1.1 or 1.2 write binary data
//if Binary then StrList.Add('%' + #210#211#212#213);
// write copyright as comment
//StrList.Add('%Producer PdfGenerator by Stefan Menten');
end;

// --- TPdfCatalog ---------------------------------------
// 11.01.2000 21:48:00
// holds the catalog information of the pdf-file
//

// --- TPdfCatalog.Write ---------------------------------------
// 11.01.2000 21:48:44
// write out the catalog information to the string-list
//
procedure TPdfCatalog.Write(var StrList: TStringList;
          const sPages, sOutline: String);
var
   wOffset: Word;
begin
wOffset := GetByteOffset(StrList);
AddObj2XREF(wOffset, Generation, 'n');
with StrList do
  begin
  WriteObjHeader(StrList, Self);
  if bOffset then Add('%Offset ' + IntToStr(wOffset));
  Add('/Pages ' + sPages);
  Add('/Outlines ' + sOutline);
  Add('>>');
  Add('endobj');
  end;
end;


// --- TPdfOutlines ---------------------------------------
// 11.01.2000 21:49:17
// holds the outline information of the pdf-file
//

// --- TPdfOutlines.Write ---------------------------------------
// 11.01.2000 21:49:21
// write out the outline information of the pdf-file
//
procedure TPdfOutlines.Write(var StrList: TStringList);
var
   wOffset: Word;
begin
wOffset := GetByteOffset(StrList);
AddObj2XREF(wOffset, Generation, 'n');
with StrList do
  begin
  WriteObjHeader(StrList, Self);
  if bOffset then Add('%Offset ' + IntToStr(wOffset));
  Add('/Count ' + IntToStr(OutlCount));
  Add('>>');
  Add('endobj');
  end;
end;


// --- TPdfPages ---------------------------------------
// 11.01.2000 21:50:39
// holds the information of all pages generated for the pdf-file
//

// --- TPdfPages.Write ---------------------------------------
// 11.01.2000 21:51:09
// write out pages object
// that are informations to all pages in the file
procedure TPdfPages.Write(var StrList: TStringList;
          const aPages: array of TPdfPage);
var
   Kids: String;
   xKids: Integer;
   wOffset: Word;
begin
wOffset := GetByteOffset(StrList);
// add the reference of the pages object
AddObj2XREF(wOffset, Generation, 'n');
// create list of references to pages
for xKids := 0 to PageCount-1 do
    Kids := Kids + GetReferenceTo(aPages[xKids]) + ' ';
// write out information to pdf-file
with StrList do
  begin
  WriteObjHeader(StrList, Self);
  if bOffset then Add('%Offset ' + IntToStr(wOffset));
  Add('/Count ' + IntToStr(PageCount));
  Add('/Kids [' + Kids + ']');
  Add('>>');
  Add('endobj');
  end;
end;

// --- TPdfContents.Create ---------------------------------------
// 11.01.2000 21:55:55
// object for the contens of 1 page
//

// --- TPdfContents.Create ---------------------------------------
// 11.01.2000 21:56:14
// create and init the object
//
constructor TPdfContents.Create;
begin
inherited create;
ConList := TStringList.Create;
BmpCount := 0;
end;

// --- TPdfContents.Write ---------------------------------------
// 11.01.2000 21:56:32
// write aut the contents
//
procedure TPdfContents.Write(var StrList: TStringList);
var
   x: Integer;
   wOffset: Word;
begin
with StrList do
  begin
  wOffset := GetByteOffset(StrList);
  AddObj2XREF(wOffset, self.Generation, 'n');
  Add(IntToStr(Self.ObjID) + ' ' + IntToStr(Self.Generation) +
      ' obj');
  if bOffset then Add('%Offset ' + IntToStr(wOffset));
  Add('<< /Length ' + IntToStr(GetByteOffset(Self.ConList)) + ' >>');
  Add('stream');
  for x := 0 to Self.ConList.Count-1 do
      Add(Self.ConList[x]);
  Add('endstream');
  Add('endobj');
end;
end;

// --- TPdfPage ---------------------------------------
// 13.01.00 20:34:34
// object holds one page of the file
//

// --- TPdfPage.Create ---------------------------------------
// 13.01.00 20:35:00
// create the object and the contents object
// and set media box to default
constructor TPdfPage.Create;
begin
inherited create;
PdfContents := TPdfContents.Create;
PdfContents.ObjType := 'Contents';
PdfContents.PdfOwner := Self.PdfOwner;
// MediaBox vorbelegen
with MediaBox do
  begin
  mbTop   := 0;
  mbLeft  := 0;
  mbHeight  := 792;
  mbWidth := 612;
  end;
end;

// --- TPdfPage.Write ---------------------------------------
// 13.01.00 20:35:37
// write out the page to the string list
//
procedure TPdfPage.Write(var StrList: TStringList);
var
   wOffset: Word;
   x, ImageRes, ContentsRes, ActImage: Integer;
   sTemp, sImagePath: String;
begin
with StrList do
  begin
  // write images
  if PdfContents.BmpCount > 0 then
    begin
    ImageRes := GetNextID;
    ContentsRes := GetNextID;

    // write the content object for each image
    //sTemp := '[ ';
    for x := 1 to PdfContents.BmpCount do
      begin
      sImagePath := IntToStr(PdfContents.Bitmaps[x].Width) + ' 0 0 -' +
                    IntToStr(PdfContents.Bitmaps[x].Height) + ' ' +
                    FloatToStrDot(PdfContents.Bitmaps[x].Left) + ' ' +
                    FloatToStrDot(PdfContents.Bitmaps[x].Top) + ' cm';

      ActImage := GetNextID;
      // add image to content resource
      sTemp := sTemp + IntToStr(ActImage) + ' 0 R ';
      // write image object
      wOffset := GetByteOffset(StrList);
      AddObj2XREF(wOffset, Generation, 'n');
      Add(IntToStr(ActImage) + ' 0 obj');
      Add('<< /Length ' + IntToStr(length(sImagePath) + 9) + ' >>');
      Add('stream');
      Add('q');
      Add(sImagePath);
      Add('/Im' + IntToStr(x) + ' Do');
      Add('Q');
      Add('endstream');
      Add('endobj');
      end;
    end;

  // Page schreiben
  wOffset := GetByteOffset(StrList);
  AddObj2XREF(wOffset, Generation, 'n');
  WriteObjHeader(StrList, Self);
  if bOffset then Add('%Offset ' + IntToStr(wOffset));
  Add('/Parent ' + GetReferenceTo(PdfOwner.FPdfBody.PdfPages));
  Add('/MediaBox [' + FloatToStrDot(PdfOwner.FPageSettings.FMediaBox.mbTop) + ' '
                    + FloatToStrDot(PdfOwner.FPageSettings.FMediaBox.mbLeft) + ' '
                    + FloatToStrDot(PdfOwner.FPageSettings.FMediaBox.mbWidth) + ' '
                    + FloatToStrDot(PdfOwner.FPageSettings.FMediaBox.mbHeight) + ']');
  // if images load add resource for that
  if PdfContents.BmpCount > 0 then
    begin
    Add('/Resources <<');
    Add('/XObject <<');
    for x := 1 to PdfContents.BmpCount do
      Add('/Im' + InttoStr(x) + ' ' + GetReferenceTo(PdfContents.Bitmaps[x]));
    Add('>>');
    Add('/Font ' + GetReferenceTo(PdfOwner.FPdfBody.PdfProcRef));
    Add('/ProcSet [ /PDF /Text /ImageC /ImageI ]');
    Add('>>');
    Add('/Contents [ ' + GetReferenceTo(Self.PdfContents) + ' ' +
        sTemp + ' ]');
    end
  else
    begin
    Add('/Resources <<');
    Add('/Font ' + GetReferenceTo(PdfOwner.FPdfBody.PdfProcRef));
    Add('/ProcSet [ /PDF /Text ]');
    Add('>>');
    Add('/Contents ' + GetReferenceTo(Self.PdfContents) + '');
    end;
  Add('>>');
  Add('endobj');

  // write contents
  PdfContents.Write(StrList);
  // write images
  for x := 1 to PdfContents.BmpCount do
    begin
    PdfContents.Bitmaps[x].Write(StrList, x);
    end;
  end;

end;

// --- TPdfProcRef ---------------------------------------
// 13.01.00 20:36:07
//
//

// --- TPdfProcRef.Write ---------------------------------------
// 13.01.00 20:37:22
// write out the default fonts
//
procedure TPdfProcRef.Write(var StrList: TStringList);
var
   wOffset: Word;
   x: Integer;
begin
wOffset := GetByteOffset(StrList);
AddObj2XREF(wOffset, Generation, 'n');
with StrList do
  begin
  Add(IntToStr(Self.ObjID) + ' ' + IntToStr(Self.Generation) + ' obj');
  if bOffset then Add('%Offset ' + IntToStr(wOffset));
  Add('<<');
  for x := 1 to 14 do
      if PdfOwner.BaseFonts[x] > '' then
         Add('/F' + IntToStr(x) + ' ' + PdfOwner.BaseFonts[x]);
  Add('>>');
  Add('endobj');
  end;
end;

// --- TPdfBody ---------------------------------------
// 13.01.00 20:37:43
// PdfBody ist the main file object
//

// --- TPdfBody.InitBody ---------------------------------------
// 13.01.00 20:38:16
// init the body and create objects
//
procedure TPdfBody.InitBody;
begin
// catalog
PdfCatalog := TPdfCatalog.create;
PdfCatalog.ObjType := 'Catalog';
PdfCatalog.PdfOwner := Self.PdfOwner;
// outlines
PdfOutlines := TPdfOutlines.create;
PdfOutlines.ObjType := 'Outlines';
PdfOutlines.OutlCount := 0;
PdfOutlines.PdfOwner := Self.PdfOwner;
// pages
PdfPages := TPdfPages.create;
PdfPages.ObjType := 'Pages';
PdfPages.PdfOwner := Self.PdfOwner;
// proc ref
PdfProcRef := TPdfProcRef.create;
PdfProcRef.ObjType := 'ProcRef';
PdfProcRef.PdfOwner := Self.PdfOwner;
end;

// --- TPdfBody.WritePdfBody ---------------------------------------
// 13.01.00 20:38:37
// write out the pdf-body
//
function TPdfBody.WritePdfBody(var StrList: TStringList;
         const iPageCount: Integer): Boolean;
var
   xPage: Integer;
begin
result := true;
try
// Catalog schreiben
PdfCatalog.Write(StrList,
   GetReferenceTo(PdfPages),
   GetReferenceTo(PdfOutlines));
// Outlines schreiben
PdfOutlines.Write(StrList);
// Pages schreiben
PdfPages.Write(StrList, PdfPage);

for xPage := 1 to iPageCount do
  PdfPage[xpage].Write(StrList);

// ProcRef schreiben
PdfProcRef.Write(StrList);
except
result := false;
end;
end;

// --- TPdfCrossRefernec.Write ---------------------------------------
// 13.01.00 20:38:59
// the cross reference holds references to all objects in file
//

// --- TPdfCrossRefernec.Write ---------------------------------------
// 13.01.00 20:39:25
// write out the proc ref
//
procedure TPdfCrossRefernec.Write(var StrList: TStringList);
var
   x: Integer;
begin
StartOffset := GetByteOffset(StrList);
XRefCount   := g_XREFList.Count+1;
with StrList do
  begin
  Add('xref');
  Add('0 ' + IntToStr(XRefCount));
  Add('0000000000 65535 f');
  for x := 0 to g_XREFList.Count-1 do
      Add(g_XREFList[x]);
  end;
end;

// --- TPdftrailer ---------------------------------------
// 13.01.00 20:39:44
// the trailer ends the pdf-file
// it holds the information of file size and root element

// --- TPdftrailer.Write ---------------------------------------
// 13.01.00 20:39:47
// write out the trailer
//
procedure TPdftrailer.Write(var StrList: TStringList);
begin
with StrList do
  begin
  Add('trailer');
  Add('<<');
  Add('/Size ' + IntToStr(PdfOwner.FPdfCrossRefernec.XRefCount));
  Add('/Root ' + GetReferenceTo(PdfOwner.FPdfBody.PdfCatalog));
  Add('>>');
  Add('startxref');
  Add(IntToStr(PdfOwner.FPdfCrossRefernec.StartOffset));
  Add('%%EOF');
  end;
end;


// --- TPdfCmp ---------------------------------------
// 13.01.00 20:40:57
// the compoents object
//

// --- TPdfCmp.Create ---------------------------------------
// 13.01.00 20:41:05
// create and init object
constructor TPdfCmp.Create(AOwner : TComponent);
begin
// first inherid original constructor
inherited Create(AOwner);
// then create nessary objects
FPdfVersion := cPdf11;
g_IDNextFree := 1;
PdfInit;
end;

// --- TPdfCmp.Destroy ---------------------------------------
// 13.01.00 20:41:39
// destroy the object
destructor TPdfCmp.Destroy;
begin
// first destroy objects
if FPdfTrailer <> NIL then FPdfTrailer.Destroy;
if FPdfCrossRefernec <> NIL then FPdfCrossRefernec.Destroy;
if FPdfBody <> NIL then FPdfBody.Destroy;
if FPdfHeader <> NIL then FPdfHeader.Destroy;
if g_XREFList <> NIL then g_XREFList.Destroy;
if FPdfStrList <> NIL then FPdfStrList.Destroy;
if FPageSettings <> NIL then FPageSettings.destroy;
// then inherid original destroctor
inherited Destroy;
end;

// --- TPdfCmp.SetMargins ---------------------------------------
// 13.01.00 20:41:51
// set the page margins
//
procedure TPdfCmp.SetMargins(const maLeft, maRight, maTop, maBottom: Integer);
begin
PageSettings.Margins(maLeft, maRight, maTop, maBottom);
// if assigned, set margins to actual page
if (FPdfActPage <> NIL) then
   begin
   FPdfActPage^.CropBox := PageSettings.FCropBox;
   FPdfActPage^.MediaBox := PageSettings.FMediaBox;
   FYPosActual := FPdfActPage^.MediaBox.mbHeight-FPdfActPage^.CropBox.mbTop;
   end;
end;

// --- TPdfCmp.SetFileNamePdf ---------------------------------------
// 13.01.00 20:42:06
// Set Filename to save PDF-File
procedure TPdfCmp.SetFileNamePdf(const Value: String);
begin
if (Value > '') and (Value <> FFileNamePdf) then
   FFileNamePdf := Value;
end;

// --- TPdfCmp.SetPdfVersion ---------------------------------------
// 13.01.00 20:42:18
// set the pdf-version for the file
procedure TPdfCmp.SetPdfVersion(const Value: String);
begin
if Value > '' then
   begin
   if Value = cPdf12 then
      FPdfVersion := cPdf12
   else if Value = cPdf11 then
           FPdfVersion := cPdf11
        else
           FPdfVersion := cPdf10;
   end
else
   FPdfVersion := cPdf10;
end;

// --- TPdfCmp.SetDefaults ---------------------------------------
// 13.01.00 20:42:36
// set the default for the output
//
procedure TPdfCmp.SetDefaults;
begin
with PdfWriteParDef do
  begin
  HeadingFont   := 'F1';
  HeadingSize   := 14;
  HeadingAlign  := taPdfLeft;
  TextFont      := 'F1';
  TextSize      := 10;
  TextAlign     := taPdfLeft;
  end;
end;

// --- TPdfCmp.GetXPosition ---------------------------------------
// 13.01.00 20:43:03
// get X position depended to align
function TPdfCmp.GetXPosition(const sText: String;
         taAlign: TPdfAlignment; const iFont, iFontSize: Integer): double;
begin
case taAlign of
  taPdfLeft:    begin
                result := FPdfActPage^.CropBox.mbLeft;
                end;
  taPdfRight:   begin
                result := FPdfActPage^.CropBox.mbLeft + FPdfActPage^.CropBox.mbWidth -
                       PdfBaseFonts.GetTextWidthUserSpace(iFont, iFontSize, sText);
                end;
  taPdfCenter:  begin
                result := FPdfActPage^.CropBox.mbLeft +
                       (FPdfActPage^.CropBox.mbWidth / 2) -
                       (PdfBaseFonts.GetTextWidthUserSpace(iFont, iFontSize, sText) / 2);
                end;
  else          begin
                result := 0;
                end;
  end;
end;

// --- TPdfCmp.PageAppend ---------------------------------------
// 13.01.00 20:43:14
// add one page and set it up
//
function TPdfCmp.PageAppend: Boolean;
begin
Result := true;
// Append page
try
Inc(FPageCount);
FPdfBody.PdfPages.PageCount := FPageCount;
FPdfBody.PdfPage[FPageCount] := TPdfPage.Create;
FPdfBody.PdfPage[FPageCount].PdfOwner := self;
FPdfBody.PdfPage[FPageCount].ObjType := 'Page';
// Pointer to actual page
FPdfActPage := @FPdfBody.PdfPage[FPageCount];
// actual Y position
FPdfActPage^.CropBox := PageSettings.FCropBox;
FPdfActPage^.MediaBox := PageSettings.FMediaBox;
FYPosActual := FPdfActPage^.MediaBox.mbHeight-FPdfActPage^.CropBox.mbTop;
except
Result := false;
end;
end;

// --- TPdfCmp.AddText ---------------------------------------
// 13.01.00 20:44:16
// add text to the actual page
//
procedure TPdfCmp.AddText(const xPos, yPos: Double; const FontPdfName: String;
          const FontSize: Byte; const PdfText: String);
var
   xPosNew: Double;
begin
// Text rechts positioniert
//xPosNew := FPageSettings.FMediaBox.mbWidth - 20 -
  //         PdfBaseFonts.GetTextWidthUserSpace(1, FontSize, PdfText);

//xPosNew := FPageSettings.FCropBox.mbLeft;
xPosNew := xPos;

with FPdfActPage.PdfContents.ConList do
  begin
  Add('BT');
  Add('/' + FontPdfName + ' ' + IntToStr(FontSize) + ' Tf');
  Add(FloatToStrDot(xPosNew) + ' ' + FloatToStrDot(yPos) + ' ' + ' Td (' + PdfText +') Tj');
  Add('ET');
  end;
end;

// --- TPdfCmp.WriteParagraph ---------------------------------------
// 13.01.00 20:45:07
// write a complete paragraph with heading
procedure TPdfCmp.WriteParagraph(const Heading, Text: String);
var
   tmpYPos: double;
begin
// write heading
if Heading > '' then
  begin
  // write heading
  tmpYPos := FYPosActual - (PdfWriteParDef.HeadingSize * c_DefFontHeight);
  // dont start the paragraph if end of page is near
  if (tmpYPos - (PdfWriteParDef.HeadingSize * c_DefFontHeight)) <
     (FPdfActPage.MediaBox.mbHeight - FPdfActPage.CropBox.mbHeight -
     FPdfActPage.CropBox.mbTop) then
     begin
     PageAppend;
     tmpYPos := FYPosActual - (PdfWriteParDef.HeadingSize * c_DefFontHeight);
     end;
  // write the text dependend to alignment
  AddText(GetXPosition(Heading,
                       PdfWriteParDef.HeadingAlign,
                       PdfBaseFonts.GetFontNr(PdfWriteParDef.HeadingFont),
                       PdfWriteParDef.HeadingSize),
          tmpYPos,
          PdfWriteParDef.HeadingFont,
          PdfWriteParDef.HeadingSize,
          Heading);
  // increment Y position
  FYPosActual := tmpYPos;
  end;
// write text
if Text > '' then
  begin
  FYPosActual := AddFloatingText(FPdfActPage.CropBox.mbLeft, FYPosActual,
                                 FPdfActPage.CropBox.mbWidth,
                                 PdfWriteParDef.TextAlign,
                                 Text);
  end;
end;

// --- TPdfCmp.AddBitmap ---------------------------------------
// 13.01.00 20:45:33
// add bitmap to the actual page
procedure TPdfCmp.AddBitmap(const Left, Top: double;
          const BitmapPath: String);
begin
with FPdfActPage.PdfContents do
  begin
  Inc(BmpCount);
  Bitmaps[BmpCount] := TPdfXImage.Create;
  Bitmaps[BmpCount].PdfOwner := Self;
  Bitmaps[BmpCount].SetBitmap((Left/cMM2Inch) * cInch2UserSpace,
                              (Top/cMM2Inch) * cInch2UserSpace,
                              BitmapPath);
  end;

end;

// --- TPdfCmp.DrawLineHor ---------------------------------------
// 13.01.00 20:45:46
// draw a horizontal line to the actual page
procedure TPdfCmp.DrawLineHor(const xPos, yPos, xLen, iLineWidth: Integer);
begin
with FPdfActPage.PdfContents.ConList do
  begin
  Add('BT');
  Add(IntToStr(iLineWidth) + ' w');
  Add(IntToStr(xPos) + ' '  + IntToStr(yPos) + ' m');
  Add(IntToStr(xPos+xLen) + ' ' + IntToStr(yPos) + ' l S');
  Add('ET');
  end;
end;

// --- TPdfCmp.DrawLineVer ---------------------------------------
// 13.01.00 20:46:31
// draw vertical line to the actual page
procedure TPdfCmp.DrawLineVer(const xPos, yPos, yLen, iLineWidth: Integer);
begin
with FPdfActPage.PdfContents.ConList do
  begin
  Add('BT');
  Add(IntToStr(iLineWidth) + ' w');
  Add(IntToStr(xPos) + ' '  + IntToStr(yPos) + ' m');
  Add(IntToStr(xPos) + ' ' + IntToStr(yPos+yLen) + ' l S');
  Add('ET');
  end;
end;

// --- TPdfCmp.DrawLine ---------------------------------------
// 13.01.00 20:47:00
// draw a line at ...
procedure TPdfCmp.DrawLine(const xTop, yTop, xBot, yBot, iLineWidth: Integer);
begin
with FPdfActPage.PdfContents.ConList do
  begin
  Add('BT');
  Add(IntToStr(iLineWidth) + ' w');
  Add(IntToStr(xTop) + ' '  + IntToStr(yTop) + ' m');
  Add(IntToStr(xBot) + ' ' + IntToStr(yBot) + ' l S');
  Add('ET');
  end;
end;

// --- TPdfCmp.DrawRect ---------------------------------------
// 13.01.00 20:47:19
// drwa a rectangel to the actual page
//
procedure TPdfCmp.DrawRect(const xTop, yTop, xBot, yBot, iLineWidth: Integer);
begin
with FPdfActPage.PdfContents.ConList do
  begin
  Add('BT');
  Add(IntToStr(iLineWidth) + ' w');
  Add(IntToStr(xTop) + ' '  + IntToStr(yTop) + ' m'); // move to
  Add(IntToStr(xBot) + ' ' + IntToStr(yTop) + ' l');  // top hor line
  Add(IntToStr(xBot) + ' ' + IntToStr(yBot) + ' l');  // rigth ver line
  Add(IntToStr(xTop) + ' ' + IntToStr(yBot) + ' l');  // bot hor line
  Add(IntToStr(xTop) + ' '  + IntToStr(yTop) + ' l S'); // back and end
  Add('ET');
  end;
end;

// --- TPdfCmp.GetTextLengthMM ---------------------------------------
// 13.01.00 20:47:37
// Get text length from Pdf base font
function TPdfCmp.GetTextLengthMM(const byFontNr, byFontSize: Byte;
         const sText: String): Word;
var
   tmpResult: double;
begin
tmpResult := PdfBaseFonts.GetTextWidthUserSpace(byFontNr, byFontSize, sText);
tmpResult := (tmpResult/cInch2UserSpace)*cMM2Inch;
Result := trunc(tmpResult);
end;

// --- TPdfCmp.SetOrientation ---------------------------------------
// 13.01.00 20:47:46
// set page orientation
procedure TPdfCmp.SetOrientation(const Orientation: TPdfOrientation);
begin
if Orientation = PdfLandscape then
   FPageSettings.FOrientation := PdfLandscape
else
   FPageSettings.FOrientation := PdfPortrait;
end;

// --- TPdfCmp.PdfSave ---------------------------------------
// 13.01.00 20:47:59
// save the pdf-file
procedure TPdfCmp.PdfSave;
begin
if not WritePdfFile then  ;
end;

// --- TPdfCmp.PdfInit ---------------------------------------
// 13.01.00 20:48:15
// init the object
procedure TPdfCmp.PdfInit;
begin
// Base fonts
PdfBaseFonts := TPdfBaseFonts.create;
// holts page settings for size, margins aso
FPageSettings := TPageSettings.create;
// stringlist to hold content
FPdfStrList := TStringList.create;
// stringlist for XREF
g_XREFList := TStringList.create;
// pdf file header
FPdfHeader := TPdfHeader.create;
FPdfHeader.ObjType := 'Header';
FPdfHeader.PdfOwner := Self;
// pdf body
FPdfBody   := TPdfBody.Create;
FPdfBody.ObjType := 'Body';
FPdfBody.PdfOwner := Self;
FPdfBody.InitBody;
// pdf cross reference
FPdfCrossRefernec := TPdfCrossRefernec.Create;
FPdfCrossRefernec.ObjType := 'XREF';
FPdfCrossRefernec.PdfOwner := Self;
// pfd trailer
FPdfTrailer := TPdfTrailer.Create;
FPdfTrailer.ObjType := 'Trailer';
FPdfTrailer.PdfOwner := Self;
// Pointer to ActPage
FPdfActPage := NIL;
// Set defaults
SetDefaults;
end;

// --- TPdfCmp.WriteFonts ---------------------------------------
// 13.01.00 20:48:32
// write base fonts out to pdf-file
procedure TPdfCmp.WriteFonts(var StrList: TStringList);

  procedure WriteFont(const FontNr: Word; const FontName: String);
  var
     wOffset, ID: Word;
  begin
  wOffset := GetByteOffset(StrList);
  AddObj2XREF(wOffset, 0, 'n');
  with StrList do
    begin
    ID := GetNextID;
    BaseFonts[FontNr] := IntToStr(ID) + ' 0 R';
    Add(IntToStr(ID) + ' 0 obj');
    Add('<<');
    if bOffset then Add('%Offset ' + IntToStr(wOffset));
    Add('/Type /Font');
    Add('/Subtype /Type1');
    Add('/Name /F' + IntToStr(FontNr));
    Add('/BaseFont /' + FontName);
    Add('/Encoding /WinAnsiEncoding');
    Add('>>');
    Add('endobj');
    end;
  end;
begin
// Write Fonts to file
WriteFont(1, 'Helvetica');
WriteFont(2, 'Helvetica-Bold');
WriteFont(3, 'Helvetica-Oblique');
WriteFont(4, 'Helvetica-BoldOblique');
WriteFont(5, 'Courier');
WriteFont(6, 'Courier-Bold');
WriteFont(7, 'Courier-Oblique');
WriteFont(8, 'Courier-BoldOblique');
WriteFont(9, 'Times-Roman');
WriteFont(10, 'Times-Bold');
WriteFont(11, 'Times-Italic');
WriteFont(12, 'Times-BoldItalic');
WriteFont(13, 'Symbol');
WriteFont(14, 'ZapfDingbats');
end;

// --- TPdfCmp.SplitString ---------------------------------------
// 13.01.00 20:48:51
// split longer strings down to 255 char
function TPdfCmp.SplitString(const sString: String): String;
begin
end;

// --- TPdfCmp.WritePdfFile ---------------------------------------
// 13.01.00 20:48:59
// write the file to disk
function TPdfCmp.WritePdfFile: Boolean;
begin
Result := true;
// clear StringList
//FPdfStrList := TStringList.create;
FPdfStrList.Clear;
// write header
FPdfHeader.Write(FPdfStrList, true, FPdfVersion);
// write fonts to file
WriteFonts(FPdfStrList);
// write body
FPdfBody.WritePdfBody(FPdfStrList, PdfPageCount);
// write XREF
FPdfCrossRefernec.Write(FPdfStrList);
// write Trailer
FPdfTrailer.Write(FPdfStrList);
// save PDF-File
if FFileNamePdf > '' then
   FPdfStrList.SaveToFile(FFileNamePdf)
else
   Result := false;
end;

// --- TPdfCmp.AddFloatingText ---------------------------------------
// 13.01.00 20:49:17
// add flow text
function TPdfCmp.AddFloatingText(const xStart, yStart, MaxWidth: double;
         const PdfAlign: TPdfAlignment; sText: String): double;
var
   tmpList, ResultList: TStringList;
   tmpYPos, tmpXPos, yOffset, dTextWidth, dLineWidth, xLastOffset, yBottom: double;
   x, y: Integer;
   tmpLine, tmpWord, sActFont: String;
   NewPage: Boolean;

   function SetPosition(const LastLine: Boolean): Boolean;
   var
      xPos, yPos, xAlignOffset: double;
   begin
   result := true;
   // set absolute y position
   tmpYPos := tmpYPos - (PdfWriteParDef.TextSize * c_DefFontHeight);
   // check if empty line
   if dLineWidth = 0 then
     begin
     // store relative y position
     yOffset :=  yOffset -(PdfWriteParDef.TextSize * c_DefFontHeight);
     result := false;
     exit;
     end;
   if xLastOffset = 0 then xLastOffset := MaxWidth;

     // set x depended to alignment
     case PdfAlign of
       taPdfLeft:    begin
                     xAlignOffset := 0;
                     end;
       taPdfRight:   begin
                     // set y relative to last y
                     xAlignOffset :=  xLastOffset - dLineWidth;
                     // store line width for next line
                     xLastOffset := dLineWidth;
                     end;
       taPdfCenter:  begin
                     // set y relative to half page width
                     xAlignOffset := (xLastOffset / 2) - (dLineWidth / 2);
                     // store line width for next line
                     xLastOffset := dLineWidth;
                     end;
       taPdfJustify: begin
                     // start at the left
                     xAlignOffset := 0;
                     // get difference between linewidth and MaxWidth
                     if LastLine then
                       FPdfActPage.PdfContents.ConList.Add('0 Tw')
                     else
                       FPdfActPage.PdfContents.ConList.Add(
                         FloatToStrDot((MaxWidth-dLineWidth)/(strTokenCount(tmpLine, ' ')-1))  + ' Tw');
                     end;
       else
       end;
     // on new page set absolute position
     if NewPage then
       begin
       xPos := xStart + xAlignOffset;
       yPos := tmpYPos;
       NewPage := false;
       end
     else
       begin
       xPos := xAlignOffset; //tmpXPos + xAlignOffset;
       yPos := yOffset;
       end;
     // write the text position
     FPdfActPage.PdfContents.ConList.Add(FloatToStrDot(xPos) + ' ' +
         FloatToStrDot(yPos) + ' TD');
     tmpXPos := 0;
     yOffset :=  -(PdfWriteParDef.TextSize * c_DefFontHeight);
   end;

begin
tmpList := NIL;
ResultList := NIL;
result := yStart;
try
// remove all #10 and #13
sText := strReplace(sText, #13, '');
sText := strReplace(sText, #10, '');
// create string lists
ResultList := TStringList.create;
tmpList := TStringList.create;
// write each row devided by <br> to tmpList
while Pos(c_Break, sText) > 0 do
  begin
  tmpList.Add( Copy(sText, 1, Pos(c_Break, sText)-1) );
  sText := Copy(sText, Pos(c_Break, sText)+Length(c_Break), Length(sText) - Pos(c_Break, sText));
  end;
// add last row
tmpList.Add(sText);
//yOffset := tmpYPos - (PdfWriteParDef.TextSize * c_DefFontHeight);
tmpXPos := XStart;
NewPage := false;
xLastOffset := 0;
// set start position for text
tmpYPos := yStart - (PdfWriteParDef.TextSize * c_DefFontHeight);
// begin text
FPdfActPage.PdfContents.ConList.Add('BT');
// set starting position
FPdfActPage.PdfContents.ConList.Add(FloatToStrDot(xStart) + ' ' +
  FloatToStrDot(tmpYPos) + ' TD');
// get bottom margin
yBottom := FPdfActPage.MediaBox.mbHeight - FPdfActPage.CropBox.mbHeight -
        FPdfActPage.CropBox.mbTop - (PdfWriteParDef.TextSize * c_DefFontHeight);

// set default font
sActFont := '/' + PdfWriteParDef.TextFont + ' ' +
         IntToStr(PdfWriteParDef.TextSize) + ' Tf';
FPdfActPage.PdfContents.ConList.Add(sActFont);

for x := 0 to tmpList.Count-1 do
  begin
  dTextWidth := 0;
  dLineWidth := 0;
  if Pos(c_FontStart, tmpList[x]) > 0 then
     begin
     // use defined font
     end
  else
     begin
     // use default font
     // process for each word in line
     for y := 0 to strTokenCount(tmpList[x], ' ')-1 do
         begin
         // get actual word
         tmpWord := strTokenAt(tmpList[x], ' ', y);
         // add blank if not first word
         if dTextWidth > 0 then tmpWord := ' ' + tmpWord;
         // store width of line
         dLineWidth := dTextWidth;
         // add lenght of line
         dTextWidth := dTextWidth +
                    PdfBaseFonts.GetTextWidthUserSpace(
                      PdfBaseFonts.GetFontNr(PdfWriteParDef.TextFont),
                      PdfWriteParDef.TextSize, tmpWord);

         // if width of actual line > max width write the text
         if dTextWidth > MaxWidth then
            begin
            // check for new page
            if tmpYPos < yBottom then
               begin
               FPdfActPage.PdfContents.ConList.Add('ET');
               PageAppend;
               FPdfActPage.PdfContents.ConList.Add('BT');
               FPdfActPage.PdfContents.ConList.Add(sActFont);
               tmpYPos := FYPosActual;
               NewPage := true;
               xLastOffset := 0;
               end;
            // Set position for text
            if SetPosition(False) then
              // show text
              FPdfActPage.PdfContents.ConList.Add('(' + Trim(tmpLine) + ') Tj');
            // start next line
            dLineWidth := 0;
            tmpLine := Trim(strTokenAt(tmpList[x], ' ', y));
            dTextWidth := PdfBaseFonts.GetTextWidthUserSpace(
                       PdfBaseFonts.GetFontNr(PdfWriteParDef.TextFont),
                       PdfWriteParDef.TextSize,
                       tmpLine );
            end
         else
           begin
           // add word to tmpLine
           tmpLine := tmpLine + tmpWord;
           end;
         end;
     // write rest of line
     dLineWidth := dTextWidth;
     if SetPosition(True) then
       begin
       // show text
       FPdfActPage.PdfContents.ConList.Add('(' + Trim(tmpLine) + ') Tj');
       end;
     tmpLine := '';
     end;
  end;
FPdfActPage.PdfContents.ConList.Add('ET');
finally
result := tmpYPos;
tmpList.Free;
ResultList.Free;
end;
end;

// ---------------------------------------------------------------------------

// Register component --------------------------------------------------------
Procedure Register;
Begin
RegisterComponents( 'PDF-Cmp', [TPdfCmp] );
End;
end.
