{ PdfGenerator Copyright 1999 by Stefan Menten

  Unit UPdfFilters
  Handling of filters for PdfGenerator }

unit UPdfFilters;

interface

uses Classes, SysUtils, math, Forms;

{Filters are used to convert 8 bit strings to 7 bit strings, so they can
 transferd to 7 bit computers to. Here Base85 is used}

type TFilter = (fiASCII85);

// arrays for base85
var Power85: array[1..5] of Double = (85*85*85*85, 85*85*85, 85*85, 85, 1);
var Power256: array[1..4] of Double = (256*256*256, 256*256, 256, 1);

// encode string of Char to base85
function EncodeBase85Char(const sUncoded: String): String;
// encode string of decimals to base85
function EncodeBase85Dec(const sUncoded: String): String;
// encode string of decimals to base85
function EncodeBase85DecFromStream(sUncoded: TMemoryStream): String;

// decode string of decimals from base85
function DecodeBase85Char(const sCoded: String): String;
// decode string of decimals from base85
function DecodeBase85Dec(const sCoded: String): String;


implementation

uses xProcs1;

// --- Encode85 ---------------------------------------
// 13.01.00 21:59:00
// change 1 char to base85
//
function Encode85(const iCoded: Comp; const iCount: Integer): String;
var
   itest, Rest: double;
begin
// return z if all char are 0
if (iCoded = 0) and (iCount = 4) then
  result := 'z'
else
  begin
  Rest := iCoded;
  itest := (Rest / IntPower(85, 4));
  result := Char(trunc(itest + 33));

  if iCount > 0 then
    begin
    Rest  := Rest - (trunc(itest) * IntPower(85, 4));
    itest := (Rest / IntPower(85, 3));
    result := result + Char(trunc(itest + 33));
    end;
 if iCount > 1 then
    begin
    Rest  := Rest - (trunc(itest) * IntPower(85, 3));
    itest := (Rest / IntPower(85, 2));
    result := result + Char(trunc(itest + 33));
    end;
 if iCount > 2 then
    begin
    Rest  := Rest - (trunc(itest) * IntPower(85, 2));
    itest := (Rest / IntPower(85, 1));
    result := result + Char(trunc(itest + 33));
    end;
  if iCount > 3 then
    begin
    Rest  := Rest - (trunc(itest) * IntPower(85, 1));
    itest := (Rest / IntPower(85, 0));
    result := result + Char(trunc(itest + 33));
    end;
  end;
end;

// --- Decode85 ---------------------------------------
// 13.01.00 21:59:49
// change 1 char from base85
//
function Decode85(const iCoded: Comp; const iCount: Integer): String;
var
   test, Rest: Double;
begin
Rest := iCoded;
test := (Rest / IntPower(256, 3));
result := IntToStr(Trunc(test)) + ' ';
if iCount > 1 then
  begin
  Rest := Rest - (trunc(test) * IntPower(256, 3));
  test := (Rest / IntPower(256, 2));
  result := result + IntToStr(Trunc(test)) + ' ';
  end;
if iCount > 2 then
  begin
  Rest := Rest - (trunc(test) * IntPower(256, 2));
  test := (Rest / IntPower(256, 1));
  result := result + IntToStr(Trunc(test)) + ' ';
  end;
if iCount > 3 then
  begin
  Rest := Rest - (trunc(test) * IntPower(256, 1));
  test := (Rest / IntPower(256, 0));
  result := result + IntToStr(Trunc(test)) + ' ';
  end;
end;

// --- EncodeBase85Char ---------------------------------------
// 13.01.00 22:00:15
// encode string of Char to base85
function EncodeBase85Char(const sUncoded: String): String;
var
   iPos, iCount: Integer;
   iCoded: Comp;

begin
// encodes a string
// formula to encode is descriped in PDFSpec page 51
iCoded := 0;
iCount := 0;
for iPos := 1 to Length(sUncoded) do
  begin
  Inc(iCount);
  iCoded := iCoded + Ord(sUncoded[iPos]) * Power256[iCount];

  // add the end sign
  if iPos = Length(sUncoded) then
    begin
    // encode last char
    if iCount > 0 then
       result := result + Encode85(iCoded, iCount);
    result := result + '~>';
    exit;
    end
  else
    if iCount = 4 then
       begin
       result := result + Encode85(iCoded, iCount);
       iCoded := 0;
       iCount := 0;
       end;
  end;
end;

// --- EncodeBase85Dec ---------------------------------------
// 13.01.00 22:00:34
// encode string of decimals to base85
// input is separeted by blank
function EncodeBase85Dec(const sUncoded: String): String;
var
   iPos, iCount, iTokenCount: Integer;
   iCoded: Comp;
   sTemp: String;
begin
// encodes a string
// formula to encode is descriped in PDFSpec page 51
iCoded := 0;
iCount := 0;
result := '';
iTokenCount := strTokenCount(sUncoded, ' ');
for iPos := 0 to iTokenCount-1 do
  begin
  sTemp := strTokenAt(sUncoded, ' ', iPos);
  try
  Inc(iCount);
  iCoded := iCoded + StrToInt(sTemp) * Power256[iCount];
  except
  end;
  // add the end sign
  if iPos = iTokenCount-1 then
    begin
    if iCount > 0 then
       result := result + Encode85(iCoded, iCount);
    result := result + '~>';
    exit;
    end
  else
    if iCount = 4 then
       begin
       result := result + Encode85(iCoded, iCount);
       iCoded := 0;
       iCount := 0;
       end;
  Application.ProcessMessages;
  end;
end;

// --- EncodeBase85DecFromStream ---------------------------------------
// 13.01.00 22:00:54
// encode string of decimals to base85
function EncodeBase85DecFromStream(sUncoded: TMemoryStream): String;
var
   iPos, iCount, iTokenCount: Integer;
   iCoded: Comp;
   sTemp: Byte;
begin
// encodes a string
// formula to encode is descriped in PDFSpec page 51
iCoded := 0;
iCount := 0;
result := '';
sUncoded.Position := 0;
iTokenCount := sUncoded.Size;
for iPos := 0 to iTokenCount-1 do
  begin
  sUncoded.Read(sTemp, 1);
  try
  Inc(iCount);
  iCoded := iCoded + sTemp * Power256[iCount];
  except
  end;
  // add the end sign
  if iPos = iTokenCount-1 then
    begin
    if iCount > 0 then
       result := result + Encode85(iCoded, iCount);
    result := result + '~>';
    exit;
    end
  else
    if iCount = 4 then
       begin
       result := result + Encode85(iCoded, iCount);
       iCoded := 0;
       iCount := 0;
       end;
  Application.ProcessMessages;
  end;
end;

// --- DecodeBase85Char ---------------------------------------
// 13.01.00 22:01:03
// decode string of char to String
function DecodeBase85Char(const sCoded: String): String;
var
   iCoded: Comp;
   iPos, iCount: Integer;
   sTemp: String;
begin
// iCount := 0;
// iPos := 0;
if sCoded > '' then
  begin
  iCoded := 0;
  iCount := 0;
  //iPos := 0;
  for iPos := 1 to Length(sCoded) do
    begin
    case sCoded[iPos] of
      'z': begin
           if iCount > 0 then
             // error
           else
             Result := Result + '0 0 0 0';
           end;
      '~': begin
           if sCoded[iPos+1] = '>' then
             if iCount > 0 then
               begin
               iCoded := iCoded + Power85[iCount];
               result := result + Decode85(iCoded, iCount-1);
               end;
           break;
           end;
      else
          begin
          Inc(iCount);
          iCoded := iCoded + ((Ord(sCoded[iPos]) - 33) * Power85[iCount]);
          if iCount = 5 then
            begin
            result := result + Decode85(iCoded, iCount-1);
            iCount := 0;
            iCoded := 0;
            end;
          end;
      end;


   end;
  end;

// umwandeln nach Char
sTemp := result;
Result := '';
while Pos(' ', sTemp) > 0 do
  begin
  Result := Result + Chr(StrToInt(Copy(sTemp, 1, Pos(' ', sTemp)-1)));
  sTemp := Copy(sTemp, Pos(' ', sTemp)+1, Length(sTemp)-Pos(' ', sTemp));
  end;

end;

// --- DecodeBase85dec ---------------------------------------
// 13.01.00 22:01:18
// decode string of char to decimals
// output is seperated by blank
function DecodeBase85dec(const sCoded: String): String;
var
   iCoded: Comp;
   iPos, iCount: Integer;

begin
// iCount := 0;
// iPos := 0;
if sCoded > '' then
  begin
  iCoded := 0;
  iCount := 0;
//  iPos := 0;
  for iPos := 1 to Length(sCoded) do
    begin
    case sCoded[iPos] of
      'z': begin
           if iCount > 0 then
             // error
           else
             Result := Result + '0 0 0 0';
           end;
      '~': begin
           if sCoded[iPos+1] = '>' then
             if iCount > 0 then
               begin
               iCoded := iCoded + Power85[iCount];
               result := result + Decode85(iCoded, iCount-1);
               end;
           exit;
           end;
      else
          begin
          if (sCoded[iPos] <> #13) and (sCoded[iPos] <> #10) then
             begin
             Inc(iCount);
             iCoded := iCoded + ((Ord(sCoded[iPos]) - 33) * Power85[iCount]);
             end;
          if iCount = 5 then
            begin
            result := result + Decode85(iCoded, iCount-1);
            iCount := 0;
            iCoded := 0;
            end;
          end;
      end;
    end;
  end;
end;

end.
